/* 
 *  OBS Master Clock Syncronization MONITOR
 *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ncurses.h>
#include <termios.h>


char buf[256];
char *bufptr=buf;
WINDOW *gps_window;
WINDOW *error_window;
WINDOW *wave_window;
static int seconds = 0;
static int first_reset = 0;
int ctl;     // fd for command fifo
int sec;     // fd for seconds fifo
int gps;     // fd for gps


/*
 *  allow program to clean up after itself :)
 */
void exitnice(void)
{
  beep();
  sleep(3);  // sleep in case there were any errormessages
  endwin();
  exit(1);
}


/*
 *  on-screen oscilloscope ...
 */
void update_waveform() {
  static int x = 0;
  static int y = 1;

  /* generate the correct waveform .. */
  if (((seconds % 5) == 0) && !(seconds == 5) && !(seconds == 10)) {
    if (y==0) {
      y=1;
      mvwprintw( wave_window, y, x, "%s", "|" );
      x++;
    } else {
      mvwprintw( wave_window, y, x, "%s", "|" );
      x++;
      y=0;
    }
  }

  /* if a minute has passed, clear window */
  if ( seconds == 0 ) {
    x=0;
    y=1;
    wclear( wave_window );
  }

  /* draw trace */
  mvwprintw( wave_window, y, x, "%s", "_" );
  x++;  // increase x position

  /* refresh window */
  wrefresh( wave_window );
}


/*
 *  writes a message in the "messages" window
 */
void write_info(char *streng) {
  wprintw(error_window, "%s", streng);
  wrefresh(error_window);
}


/* 
 *  "Parse" NMEA sentences
 *  NMEA_code     : NMEA sentence to look for. Returns NULL if not equal
 *  NMEA_sentence : the string to parse (will NOT be modified)
 *  NMEA_field_no : specifies which field in the sentence to return
 *  s             : if not NULL, the returned string will also be copied into this location
 */
char *NMEA_parse(char *NMEA_code, char *NMEA_sentence, int NMEA_field_no, char *s)
{
  int i;
  static char *NMEA;
  static char code[255];
  static char *codeptr = code;
  static char sent[255];
  static char *sentptr;

  /* make a copy of all strings, so we don't mess anything up */
  strcpy(code, (char *)NMEA_code);
  codeptr = code;
  strcpy(sent, (char *)NMEA_sentence);
  sentptr = sent;

  /* check to see if the sentence is the one we're looking for ..
   * return NULL if false. */
  NMEA = (char *)strsep(&sentptr, ",");
  if(strcmp(NMEA, codeptr) != 0)
    return(NULL);

  /* search the sentence for the correct fieldno..
   * if field doesn't exist, NULL will be returned */
  sentptr = (char *)strsep(&sentptr, "*");
  for(i = 1; i < (NMEA_field_no + 1); i++)
    NMEA = (char *)strsep(&sentptr, ",");

  // printf("NMEA: %s\n", NMEA);
  /* copy field into location of s if s is not NULL */
  if(s)  
    strcpy((char *)s, (char *)NMEA);

  /* return */
  return(NMEA);
}


/*
 *  oppdater klokke og sekundteller
 */
void check_NMEA(char *nmeastr) {
  //  int gps_valid = 1;
  char *tid;

  /* output the GPRMC sentence */
  if ( NMEA_parse( "$GPRMC", buf, 1, NULL ) ) {
    /* get GPS time (first field in GPRMC) */
    tid = NMEA_parse("$GPRMC", buf, 1, NULL);
    wprintw(gps_window, "%s\n", nmeastr);     // output "raw" NMEA sentence
    mvprintw(2, 52, "%s", tid);               // output GPS time
    /* first time we round a minute, syncronize the seconds counter to the clock */
    if(!first_reset) {
      if((*(tid+4) == '0') && (*(tid+5) == '0')) {
	write_info("00 seconds - sending reset signal to OBSMCS!\n");
	first_reset = 1;
	write(ctl, &first_reset, sizeof(first_reset));
      }
    }
    /* refresh window */
    wrefresh(gps_window);
    refresh();
  }
}


/*
 *  main program
 */
int main() {
  int msg=0, retval, n;
  char ch;
  fd_set rfds;
  struct timeval tv;
  static int send_seconds_flag=1;
  struct termios attr;

  /* initialize screen */
  initscr();

  /* open /dev/rtf2  (command fifo) */
  if((ctl = open("/dev/rtf2", O_WRONLY)) < 0) {
    printf("err: could not open /dev/rtf2.\n\fIs obsmcs.o loaded?\n\f");
    exitnice();
  }

  /* open /dev/rtf3  (seconds fifo) */
  if((sec = open("/dev/rtf3", O_RDONLY)) < 0) {
    printf("err: could not open /dev/rtf3.\nIs obsmcs.o loaded?\n");
    close(ctl);
    exitnice();
  }

  /* open /dev/ttyS0 (serial port 1) */
  if((gps = open("/dev/ttyS0", O_RDONLY)) < 0) {
    printf("err: could not open /dev/ttyS0\n");
    close(ctl);
    close(sec);
    exitnice();
  }

  /* configure /dev/ttyS0 */
  tcgetattr(gps, &attr);
  cfsetispeed(&attr, B4800);
  cfsetospeed(&attr, B4800);
  tcsetattr(gps, TCSANOW, &attr);

  /* write "send seconds" command to obsmcs.o */
  msg = 2;
  if(write(ctl, &msg, sizeof(msg)) < 0) {
    printf("err: could not write to /dev/rtf2\n");
    exitnice();
  }

  attron(A_BOLD);

  move(0, 0);
  printw("%s", "OBS Master Clock Syncronization Monitor");

  move(2, 0);
  printw("%s", "internal seconds counter:                 GPS clock:");

  move(4, 0);
  printw("%s", "GPS transmitted sentences:");

  move(10, 0);
  printw("%s", "Messages:");

  move(17, 0);
  printw("%s", "Waveform:");

  attroff(A_BOLD);

  move(21, 0);
  printw("%s", "-------------------------------------------------------------------------------\n");
  printw("%s", "commands:\t");
  printw("%s", "s = toggle \"send seconds\"\n");
  printw("%s", "\t\tr = reset internal second counter\n");
  printw("%s", "\t\tq = quit\n");
  move(22, 75);
  printw("%s", "TS03");
  refresh();

  gps_window = newwin(5, 80, 5, 0);
  scrollok(gps_window, TRUE);
  error_window = newwin(5, 80, 11, 0);
  scrollok(error_window, TRUE);
  wave_window = newwin(3, 80, 18, 0);
  scrollok(wave_window, FALSE);
  mvwprintw(gps_window, 0, 0, "%s", "-- none yet --\n");
  mvwprintw(error_window, 0, 0, "%s", "waiting for GPS clock seconds to become 00 ...\n");
  wrefresh(gps_window);
  wrefresh(error_window);

  /*
   *  main loop
   */
  while(1) {
    FD_ZERO(&rfds);
    FD_SET(sec, &rfds);
    FD_SET(0, &rfds);
    FD_SET(gps, &rfds);
    tv.tv_sec = 1;
    tv.tv_usec = 0;

    /* check for data from stdin, gps and obsmcs
     * timeout after one second */
    retval = select(FD_SETSIZE, &rfds, NULL, NULL, &tv);
    if(retval > 0) {  // did we get some data??

      if(FD_ISSET(sec, &rfds)) {       // got some new seconds
	/* read second counter from obsmcs */
	n = read(sec, &msg, sizeof(msg));
	if(n < 0) {
	  printw("%s", "err: could not read from /dev/rtf3\n");
	  close(ctl);
	  close(sec);
	  exit(1);
	}
	/* print second counter to screen and update waveform */
	move(2, 25);
	seconds = msg;
	printw("%02d", seconds);
	refresh();
	update_waveform();
      }

      if(FD_ISSET(0, &rfds)) {        // user pressed some keys...
	n = read(0, &ch, sizeof(ch));
	switch(ch) {
	case 'q':
	  /* write "stop sending seconds" command */
	  msg = 3;
	  write(ctl, &msg, sizeof(msg));
	  close(ctl);
	  close(sec);
	  exitnice();
	  break;
	case 's':
	  /* toggle "send seconds" */
	  send_seconds_flag ^= 1;
	  if(send_seconds_flag)
	    msg = 2;
	  else
	    msg = 3;
	  write(ctl, &msg, sizeof(msg));
	  break;
	case 'r':
	  /* reset internal second counter */
	  write_info("waiting for GPS clock seconds to become 00 ...\n");
	  first_reset = 0;
	  break;
	default:
	  break;
	}
      }

      if(FD_ISSET(gps, &rfds)) {   // gps data ready
	n = read(gps, &ch, sizeof(ch));
	*bufptr = ch;
	bufptr++;
	if(ch == '\n') {
	  *(--bufptr) = '\0';
	  bufptr = buf;
	  check_NMEA(bufptr);
	  memset(&buf, '\0', sizeof(buf));
	  bufptr = buf;
	}
      }
    }
  }
}
